/** COPYRIGHT 2021 BECK GMBH & CO. ELEKTRONIK BAUELEMENTE KG
  * 
  * THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS  OR
  * IMPLIED, INCLUDING BUT  NOT LIMITED  TO THE  WARRANTIES OF  MERCHANTABILITY,
  * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL  THE
  * AUTHORS OR  COPYRIGHT HOLDERS  BE LIABLE  FOR ANY  CLAIM, DAMAGES  OR  OTHER
  * LIABILITY, WHETHER  IN AN  ACTION OF  CONTRACT, TORT  OR OTHERWISE,  ARISING
  * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
  * IN THE SOFTWARE. */

#include <iostream>
#include <cstdio>
#include <fstream>
#include <string>
#include <regex>
#include <unistd.h>

#include "include/ejump_tcon.h"
#include "include/epd.h"
#include "include/usb_epd.h"

/** @brief      Named enum class with all possible return values. */
enum class ReturnValue {
    /** @brief      Return code for indicating successful function execution. */
    OK,

    /** @brief      Return code for indicating that a given input string could
      *             not be parsed due to it not satisfying the line format. */
    INVALID_LINE_FORMAT,
};

/** @brief      Extracts all the information from the given line of a
  *             playlist file.
  * @details    A playlist file consists of several lines with each line
  *             representing an image to show. All lines must have the following
  *             structure to be valid:
  *             - The line has to start with an arbitrary file path, allowing
  *               random characters in it (including spaces, folder separators,
  *               etc.). The file path will be treated relative to the root
  *               image file path given as the first command line parameter to
  *               this program (IMAGES_FOLDER).
  *             - After that, there may be a random number of white space
  *               characters (i.e. it doesn't matter whether you use spaces or
  *               tabs for separation, nor how many of them you're using).
  *             - Then, there has to be a number (with 5 digits max)
  *               representing the time (in seconds) to pass by before showing
  *               the next image.
  *             - Then again, there may be an arbitrary number of white space
  *               characters.
  *             - Finally, there has to be a single digit, representing the
  *               drawing mode. For more information about the available modes
  *               and their meaning, refer to the "Modes.pdf" being located in
  *               the same folder as this file.
  *             - The line may be concluded by an arbitrary number of trailing
  *               white space characters.
  *             
  *             According to this pattern, valid lines may look like
  *             
  *                 test.png               5   2
  *                 cats and dogs.jpeg     10  2
  *             
  *             In the previous examples, the two files will both be displayed
  *             with mode 2, the first one for 5 seconds, the second for 10. If
  *             the first parameter to this program (IMAGES_FOLDER) was pictures,
  *             the program will search for the two files given above (as well
  *             as for the playlist file containing the lines above) in that
  *             folder.
  * @param      line String containing the line to be parsed.
  * @param      file_name Reference to a string which content will be replaced
  *             with the file path extracted from the input line.
  * @param      duration Reference to an integer which content will be replaced
  *             with the duration value extracted from the input line.
  * @param      mode Reference to an integer which content will be replaced
  *             with the drawing mode extracted from the input line. For further
  *             information about all the modes, please read "Modes.pdf".
  * @note       If the given input line does not satisfy the pattern, none of
  *             the given variables will be changed.
  * @returns    @ref ReturnValue::OK on success,
  *             @ref ReturnValue::INVALID_LINE_FORMAT if given input line does
  *             not satisfy the pattern. */
ReturnValue parse_line(std::string const line,
                       std::string &file_name,
                       int &duration,
                       int &mode) {
    // Regular expression to be fulfilled by every line. Not only checks whether
    // the line fulfills the pattern but also captures its most important parts
    // (i.e. file path, duration and drawing mode).
    // 
    // For more information about regular expression, visit
    // https://www.regular-expressions.info/tutorialcnt.html
    std::regex pattern{"(.+?)\\s+(\\d{1,5})\\s+(\\d)\\s*"};
    std::smatch match_result;
    
    // If return value of regex_match is not zero, the given input text 'line'
    // satisfies the pattern and the match results are saved in the variable
    // 'match_result'.
    if(std::regex_match(line, match_result, pattern)) {
        file_name = match_result[1];
        duration = std::stoi(match_result[2]);
        mode = std::stoi(match_result[3]);

        return ReturnValue::OK;
    } else {
        return ReturnValue::INVALID_LINE_FORMAT;
    }
}

int main(int argc, char *argv[]) {
    if(argc != 3) {
        std::cout << "USAGE       main IMAGES_FOLDER PLAYLIST_FILE" << std::endl;
        std::cout << "       with IMAGES_FOLDER being the folder which contains all the images" << std::endl;
        std::cout << "            (the path will be relative to the location of the executable)" << std::endl;
        std::cout << "        and PLAYLIST_FILE being a path to a text file containg information" << std::endl;
        std::cout << "            about the images to be shown. The path given by PLAYLIST_FILE" << std::endl;
        std::cout << "            will be interpreted relative to IMAGES_FOLDER." << std::endl;
        std::cout << "EXAMPLE     main pictures playlist.txt" << std::endl;
        std::cout << "       with pictures being a folder located next to the executable" << std::endl;
        std::cout << "        and playlist.txt being a file within that folder." << std::endl;
        std::cout << "NOTE        Refer to in program documentation of function parse_line for" << std::endl;
        std::cout << "            information about the structure of the playlist file." << std::endl;

        return 0;
    }

    // Path prefix for all files
    std::string const path_prefix = std::string{argv[1]} + "/";

    std::cout << "Opening playlist (" << argv[2] << ")..." << std::endl;

    // Stream to the PLAYLIST_FILE
    std::ifstream file{path_prefix + std::string{argv[2]}};

    if(!file.is_open()) {
        std::cout << "Error: Could not open playlist (" << argv[2] << ")!" << std::endl;

        return 1;
    }

    std::cout << "Initialize USB communication..." << std::endl;
    usb_init();

    std::cout << "Creating interface object for communicating with e-paper display" << std::endl;
    epd *device = new epd(ED133UT2, 1); // Second parameter: Position of hardware ID-switch

    // Index of the image currently being shown
    int idx = 1;

    while(true) {
        std::string line;

        // There is no more line in the file, so restart by resetting the stream
        if(!std::getline(file, line)) {
            std::cout << std::endl << "Reached end of playlist. Restart from first line." << std::endl;
            file.clear();
            file.seekg(0, std::ios::beg);
            idx = 1;

            continue;
        }

        std::cout << std::endl << "Procedure for image #" << idx << ":" << std::endl;

        std::cout << "- Parsing line #" << idx << " of " << argv[2] << "..." << std::endl;
        std::string file_name;
        int duration;
        int drawing_mode;
        switch(parse_line(line, file_name, duration, drawing_mode)) {
            case ReturnValue::OK:
                std::cout << "- Clearing screen..." << std::endl;
                device->clrScreen();

                std::cout << "- Transferring image (" << file_name << ") to TCON board..." << std::endl;
                if(auto retval = device->ldImg2Buf((char*)((path_prefix + file_name).c_str())); retval == 101) {
                    std::cout << "- Error: Could not load image #" << idx << " (" << file_name << ")!" << std::endl;
                } else {
                    std::cout << "- Successfully transferred image (return value: " << retval << ")." << std::endl;
                    std::cout << "- Displaying image with drawing mode " << drawing_mode << "..." << std::endl;
                    device->display(drawing_mode);

                    std::cout << "- Waiting for " << duration << " seconds..." << std::endl;
                    sleep(duration);
                }
                break;

            case ReturnValue::INVALID_LINE_FORMAT:
                std::cout << "- Error: Line #" << idx << " has invalid format!" << std::endl;
                break;
        }
        ++idx;
    }

    std::cout << "Shutting down USB communication..." << std::endl;
    usb_close();

    return 0;
}
